#!/bin/sh
# -*- TCL -*-
# the next line restarts using tclsh \
exec tclsh8.3 "$0" ${1+"$@"}

#
# Where am I?
#

if {[catch {::freewrap::iswrapped $::argv0}] || \
	[::freewrap::iswrapped $::argv0] == 0} {
    #
    # Not wrapped.
    #
    set homedir [file dirname [info script]]
    set selfdir $homedir
} else {
    #
    # Probably wrapped. Try to find my home directory using $argv0
    #
    set homedir [file join [pwd] [file dirname $::argv0]]
    set selfdir [file dirname [info script]]
}

#
# Load Packages
#

foreach package {
    combat-0.7
    tcldom-2.0
    tclxml-2.0
    tcl++
} {
    if {[file exists [file join $::selfdir lib $package]]} {
	lappend auto_path [file join $::selfdir lib $package]
    } elseif {[file exists [file join $::homedir lib $package]]} {
	lappend auto_path [file join $::homedir lib $package]
    } else {
	# file exists does not seem to work in freewrap
	lappend auto_path [file join $::selfdir lib $package]
    }
}

package require combat
package require dom

#
# create Ass namespace
#

namespace eval Ass {
    variable nodecounter 0
    variable packdata
    variable homedata
    variable compdata
    variable tempfiles [list]
    variable tempidx 0
    variable tempdir ""
    variable status ""
    variable setup

    if {[file exists /tmp] && [file isdirectory /tmp]} {
	set tempdir /tmp
    } else {
	set tempdir [pwd]
    }

    set setup(host) "<auto>"
    set setup(port) 1235
    set setup(unzip) "unzip"

    set packdata(allpackages) [list]
    set homedata(allhomes) [list]
    set compdata(allcomponents) [list]
}

#
# Source support files
#

if {[file exists [file join $::selfdir setup.tcl]]} {
    source [file join $::selfdir setup.tcl]
}

if {[file exists [file join $::homedir setup.tcl]]} {
    source [file join $::homedir setup.tcl]
}

source [file join $::selfdir assio.tcl]
source [file join $::selfdir fpsfhttpd.tcl]

#
# ======================================================================
# Main
# ======================================================================
#

#
# Initialize Combat
#

set argv [eval corba::init $argv]
source [file join [file dirname [info script]] lib CCM.tcl]
combat::ir add $_ir_CCM

#
# Handle Options
#

proc usage {} {
    global argv0
    puts stderr "usage: $argv0 \[options\] <archive>"
    puts stderr "  available options:"
    puts stderr "    --dont-call-configuration-complete"
    puts stderr "    --host <host>\[:port\]"
    puts stderr "    --ccmd <IOR>"
    puts stderr "    --token <token>"
    puts stderr "    --collocated"
    puts stderr "    --autoexit"
    puts stderr "    --ior <file>"
    puts stderr "    -v"
    exit 1
}

set archive ""
set ccmdior ""
set ccmdhost ""
set verbose 0
set collocated 0
set callcc 1
set token ""
set iorfile ""
set autoexit 0

for {set i 0} {$i < [llength $argv]} {incr i} {
    switch -glob -- [lindex $argv $i] {
	--dont-call-configuration-complete {
	    set callcc 0
	}
	--collocate* -
	--colocate* {
	    incr collocated
	}
	--ccmd {
	    set ccmdior [lindex $argv [incr i]]
	}
	--host {
	    set ccmdhost [lindex $argv [incr i]]
	}
	--port {
	    set value [lindex $argv [incr i]]
	    if {[string first : $value]} {
		set ::Ass::setup(host) [lindex [split $value :] 0]
		set ::Ass::setup(port) [lindex [split $value :] 1]
	    } else {
		set ::Ass::setup(port) $value
	    }
	}
	--token {
	    set token [lindex $argv [incr i]]
	}
	--ior {
	    set iorfile [lindex $argv [incr i]]
	}
	--autoexit {
	    set autoexit 1
	}
	-v {
	    incr verbose
	}
	--* {
	    usage
	}
	default {
	    set archive [lindex $argv $i]
	}
    }
}

if {$archive == ""} {
    usage
}

#
# Display info from XML parsers
#

proc StatusDisplay {name1 name2 op} {
    global verbose
    global oldstatus

    if {$verbose && $::Ass::status != $oldstatus} {
	set oldstatus $::Ass::status
	if {$::token != ""} {
	    puts -nonewline "Assembly # $::token: "
	}
	puts $::Ass::status
    }
}

set oldstatus ""
trace variable ::Ass::status w StatusDisplay

#
# Download by http if necessary
#

if {[catch {set filename [Ass::DownloadFile $archive]} err]} {
    puts stderr "error: archive $archive not found: $err"
    exit 1
}

#
# Contact MicoCCM Daemon
#

if {$ccmdior == ""} {
    if {$ccmdhost == ""} {
	set ccmdhost "localhost:1234"
    } elseif {[string first : $ccmdhost] == -1} {
	append ccmdhost ":1234"
    }
    set ccmdior "corbaloc::$ccmdhost/MicoCCMD"
}

corba::try {
    set ccmd [corba::string_to_object $ccmdior]
} catch {... err} {
    puts stderr "error: cannot read IOR $ccmdior: $err"
    Ass::CleanupTempFiles
    exit 1
}

corba::try {
    set ia1 [$ccmd _is_a IDL:omg.org/Components/ComponentInstallation:1.0]
    set ia2 [$ccmd _is_a IDL:omg.org/Components/ServerActivator:1.0]
    set ia3 [$ccmd _is_a IDL:mico.org/MICOCCM/MicoCCMD:1.0]
} catch {... err} {
    puts stderr "error: cannot contact CCM Daemon: $err"
    Ass::CleanupTempFiles

    if {$verbose} {
	puts stderr $errorInfo
    }

    exit 1
}

if {!$ia1 || !$ia2} {
    puts stderr "error: ccmd object does not look like MicoCCM Daemon"
    Ass::CleanupTempFiles

    if {$verbose} {
	puts stderr $errorInfo
    }

    exit 1
}

#
# Create Assembly object
#

if {$verbose} {
    if {$::token != ""} {
	puts -nonewline "Assembly # $::token: "
    }
    puts "Creating Assembly ..."
}

class Assembly_impl {
    private variable state "INACTIVE"

    public method _Interface {} {
	return "IDL:omg.org/Components/Assembly:1.0"
    }

    public method build {} {
	if {$state == "INSERVICE"} {
	    return
	}

	if {$::verbose} {
	    if {$::token != ""} {
		puts -nonewline "Assembly # $::token: "
	    }
	    puts "Building Assembly ..."
	}

	#
	# Read Assembly
	#

	corba::try {
	    Ass::FileReadAssembly $::filename
	} catch {... err} {
	    if {$::verbose} {
		if {$::token != ""} {
		    puts -nonewline "Assembly # $::token: "
		}
	    
		puts "building Assembly failed: $err"
		puts $::errorInfo
	    }
	    
	    corba::throw IDL:omg.org/Components/CreateFailure:1.0
	}

	#
	# Deploy Assembly
	#

	corba::try {
	    ::Ass::DeployAssembly $::ccmd $::collocated $::callcc
	} catch {... err} {
	    if {$::verbose} {
		if {$::token != ""} {
		    puts -nonewline "Assembly # $::token: "
		}
	    
		puts "building Assembly failed: $err"
		puts $::errorInfo
	    }

	    corba::throw IDL:omg.org/Components/CreateFailure:1.0
	}

	#
	# temp files not needed any longer
	#

	Ass::CleanupTempFiles
	set state "INSERVICE"
    }

    public method tear_down {} {
	if {$state != "INSERVICE"} {
	    return
	}

	set state "INACTIVE"

	if {$::verbose} {
	    if {$::token != ""} {
		puts -nonewline "Assembly # $::token: "
	    }
	    puts "Tearing Down Assembly ..."
	}

	corba::try {
	    ::Ass::TearDownAssembly $::ccmd
	} catch {... err} {
	    if {$::verbose} {
		if {$::token != ""} {
		    puts -nonewline "Assembly # $::token: "
		}
	    
		puts "tearing down Assembly failed: $err"
		puts $::errorInfo
	    }

	    corba::throw IDL:omg.org/Components/RemoveFailure:1.0
	}
	
	Ass::CleanupTempFiles

	if {$::autoexit} {
	    set ::forever 1
	}
    }

    public method get_state {} {
	return $state
    }
}

set poa [corba::resolve_initial_references RootPOA]
set mgr [$poa the_POAManager]
set srv [Assembly_impl #auto]
set oid [$poa activate_object $srv]
set ref [$poa id_to_reference $oid]
set str [corba::object_to_string $ref]

if {$iorfile == "-"} {
    puts $str
} elseif {$iorfile != ""} {
    set reffile [open $iorfile w]
    puts -nonewline $reffile $str
    close $reffile
}

#
# Activate the POAs
#

$mgr activate

#
# If there is a token, call back the AssemblyFactory
#

if {$token != ""} {
    $ccmd callback $token $ref
}

#
# .. and serve ...
#

vwait forever

#
# flush event loop
#

update

#
# clean up
#

Ass::CleanupTempFiles
