# -*- TCL -*-
#
# ======================================================================
# Convenience functions
# ======================================================================
#

#
# tk_exec
#

proc tk_exec_fileevent {id} {
    global tk_exec_data
    global tk_exec_cond
    global tk_exec_pipe

    if {[eof $tk_exec_pipe($id)]} {
	fileevent $tk_exec_pipe($id) readable ""
	set tk_exec_cond($id) 1
	return
    }

    append tk_exec_data($id) [read $tk_exec_pipe($id) 1024]
}

proc tk_exec {args} {
    global tk_exec_id
    global tk_exec_data
    global tk_exec_cond
    global tk_exec_pipe
    global tcl_platform
    global env

    if {![info exists tk_exec_id]} {
	set tk_exec_id 0
    } else {
	incr tk_exec_id
    }

    set keepnewline 0

    for {set i 0} {$i < [llength $args]} {incr i} {
	set arg [lindex $args $i]
	switch -glob -- $arg {
	    -keepnewline {
		set keepnewline 1
	    }
	    -- {
		incr i
		break
	    }
	    -* {
		error "unknown option: $arg"
	    }
	    * {
		break
	    }
	}
    }

    if {$i > 0} {
	set args [lrange $args $i end]
    }

    #
    # hmm, seems to work better without
    #

    if {0} {
	if {$tcl_platform(platform) == "windows" && \
		[info exists env(COMSPEC)]} {
	    set args [linsert $args 0 $env(COMSPEC) "/c"]
	}
    }

    set pipe [open "|$args" r]

    set tk_exec_pipe($tk_exec_id) $pipe
    set tk_exec_data($tk_exec_id) ""
    set tk_exec_cond($tk_exec_id) 0

    fconfigure $pipe -blocking 0
    fileevent $pipe readable "tk_exec_fileevent $tk_exec_id"

    vwait tk_exec_cond($tk_exec_id)

    if {$keepnewline} {
	set data $tk_exec_data($tk_exec_id)
    } else {
	set data [string trimright $tk_exec_data($tk_exec_id) \n]
    }

    unset tk_exec_pipe($tk_exec_id)
    unset tk_exec_data($tk_exec_id)
    unset tk_exec_cond($tk_exec_id)

    if {[catch {close $pipe} err]} {
	error "pipe error: $err"
    }

    return $data
}

#
# find a file in some directories
#

proc Ass::LocateFile {filename dirname} {
    set tail [file tail $filename]

    if {[file exists [file join $dirname $filename]]} {
	return [file join $dirname $filename]
    }

    if {[file exists [file join $dirname meta-inf $filename]]} {
	return [file join $dirname meta-inf $filename]
    }

    if {[file exists [file join $dirname $tail]]} {
	return [file join $dirname $tail]
    }

    if {[file exists [file join $dirname meta-inf $tail]]} {
	return [file join $dirname meta-inf $tail]
    }

    if {[file exists [file join [file dirname $dirname] $tail]]} {
	return [file join [file dirname $dirname] $tail]
    }

    if {[file exists $filename]} {
	if {[file pathtype $filename] == "relative"} {
	    return [file join [pwd] $filename]
	}
	return $filename
    }

    if {[file exists $tail]} {
	return [file join [pwd] $filename]
    }

    error "file not found: $filename"
}

#
# find particular child nodes for a given XML node
#

proc Ass::DOMGetElement {node type} {
    set child [::dom::node cget $node -firstChild]

    while {$child != ""} {
        if {[::dom::node cget $child -nodeType] == "element" && \
                [::dom::node cget $child -nodeName] == $type} {
            return $child
        }
        set child [::dom::node cget $child -nextSibling]
    }

    error "no $type element found in [::dom::node cget $node -nodeType]"
}

proc Ass::DOMGetAll {node type} {
    set child [::dom::node cget $node -firstChild]

    set allelements [list]

    while {$child != ""} {
        if {[::dom::node cget $child -nodeType] == "element" && \
                [::dom::node cget $child -nodeName] == $type} {
            lappend allelements $child
        }
        set child [::dom::node cget $child -nextSibling]
    }

    return $allelements
}

#
# Make temp files or temp directories
#

proc Ass::MkTempFile {} {
    variable tempfiles
    variable tempidx
    variable tempdir

    set tempfilename "ass.tmp."
    append tempfilename [pid] "." [incr tempidx]
    set tempfile [file join $tempdir $tempfilename]

    lappend tempfiles $tempfile
    return $tempfile
}

proc Ass::MkTempDir {} {
    variable tempfiles
    variable tempidx
    variable tempdir

    set tempfilename "ass.tmp."
    append tempfilename [pid] "." [incr tempidx]
    set tempfile [file join $tempdir $tempfilename]

    file mkdir $tempfile

    lappend tempfiles $tempfile
    return $tempfile
}

proc Ass::CleanupTempFiles {} {
    variable tempfiles
    variable status

    foreach tempfile $::Ass::tempfiles {
	if {[file exists $tempfile]} {
	    set status "Removing temp file $tempfile"
	    update
	    catch {file delete -force $tempfile}
	}
    }

    set tempfiles [list]
}

#
# Download file into a local temp file
#

proc Ass::DownloadByHttp {url {localfilename ""}} {
    package require http
    variable tempfiles
    variable tempidx
    variable tempdir
    variable status

    set status "Downloading $url ..."
    update
    
    set hostend [string first / $url 7]
    incr hostend
    set urlfile [string range $url $hostend end]

    if {$localfilename == ""} {
	set localname "ass.tmp."
	append localname [pid] "." [incr tempidx] "."
	append localname [file tail $urlfile]
	set localfile [file join $tempdir $localname]
    } else {
	set localfile $localfilename
    }

    set out [open $localfile w]
    fconfigure $out -translation binary

    corba::try {
	set token [::http::geturl $url -channel $out]
	::http::wait $token
	if {[::http::status $token] != "ok"} {
	    error "status is not ok: [::http::status $token]"
	}
	::http::cleanup $token
	close $out
    } catch {... err} {
	catch {::http::cleanup $token}
	catch {close $out}
	catch {file delete -force $localfile}
	error "could not download $url: $err"
    }

    if {$localfilename == ""} {
	lappend tempfiles $localfile
    }

    return $localfile
}

proc Ass::DownloadFile {url {localfilename ""}} {
    #
    # Is the URL a filename?
    #
    
    if {[file exists $url]} {
	if {$localfilename == ""} {
	    if {[file pathtype $url] == "relative"} {
		return [file join [pwd] $url]
	    }
	    return $url
	}
	if {$url != $localfilename} {
	    file copy $url $localfilename
	}
	return $localfilename
    } elseif {[string range $url 0 6] == "file://"} {
	set urlfilename [string range $url 7 end]
	if {$localfilename == ""} {
	    return $urlfilename
	}
	if {$url != $urlfilename} {
	    file copy $urlfilename $localfilename
	}
	return $localfilename
    } elseif {[string range $url 0 6] == "http://"} {
	return [DownloadByHttp $url $localfilename]
    }
    
    error "could not locate file $url"
}

#
# Unzip file into a temp directory
#

proc Ass::UnzipFile {filename} {
    variable tempfiles
    variable tempidx
    variable tempdir
    variable status
    variable setup

    set localname [file tail $filename]
    set status "Unpacking archive $localname ..."
    update

    set tailroot "ass.tmp."
    append tailroot [pid] "." [incr tempidx] "."
    append tailroot [file rootname [file tail $filename]]
    set tempname [file join $tempdir $tailroot]
    
    if {[catch {tk_exec $setup(unzip) -q -n -d $tempname $filename} err]} {
	catch {file delete -force $tempname}
	set status "Error: cannot unzip $localname: $err"
	error "cannot unzip $localname: $err"
    }

    lappend tempfiles $tempname
    return $tempname
}

#
# ======================================================================
# Make new package/home/component nodes
# ======================================================================
#

#
# Packages
#

proc Ass::PackAddPackage {name type pkgurl} {
    variable nodecounter
    variable packdata

    set packnode [incr nodecounter]
    lappend packdata(allpackages) $packnode
    set packdata($packnode,id) $name
    set packdata($packnode,nodetype) $type
    set packdata($packnode,pkgurl) $pkgurl
    set packdata($packnode,children) [list]
    set packdata($packnode,refs) 0

    return $packnode
}

#
# Homes
#

proc Ass::HomeAddHome {packnode homeid} {
    variable nodecounter
    variable homedata
    variable packdata
    variable irdata

    #
    # Ref package
    #

    set packagenode $packdata($packnode,parent)
    incr packdata($packagenode,refs)

    #
    # Add a new node to the Homes tree, with subnodes for each attribute
    #

    set homenode [incr nodecounter]
    lappend homedata(allhomes) $homenode
    set homedata($homenode,nodetype) Home
    set homedata($homenode,introspected) 0
    set homedata($homenode,package) $packnode
    set homedata($homenode,children) [list]
    set homedata($homenode,refs) 0

    set homedata($homenode,id) $homeid
    set homedata($homenode,nsreg) 0
    set homedata($homenode,nsname) ""

    return $homenode
}

#
# Components
#

proc Ass::CompAddComponent {homenode compid} {
    variable nodecounter
    variable homedata
    variable compdata
    variable irdata

    #
    # Ref home
    #

    incr homedata($homenode,refs)

    #
    # Add a new node to the Components tree, with subnodes for each
    # attribute and port
    #

    set compnode [incr nodecounter]
    lappend compdata(allcomponents) $compnode
    set compdata($compnode,nodetype) Component
    set compdata($compnode,introspected) 0
    set compdata($compnode,home) $homenode
    set compdata($compnode,children) [list]
    set compdata($compnode,refs) 0

    set compdata($compnode,id) $compid
    set compdata($compnode,ccc) 1
    set compdata($compnode,nsreg) 0
    set compdata($compnode,nsname) ""

    return $compnode
}

#
# ======================================================================
# Read Assembly
# ======================================================================
#

proc Ass::FileReadAssembly {url} {
    variable packdata
    variable homedata
    variable compdata
    variable idtonode
    variable status

    set assemblyarchive [DownloadFile $url]

    set lastidx [string last / $url]
    incr lastidx
    set localname [string range $url $lastidx end]

    if {[file extension $localname] == ".aar"} {
	if {[catch {set tempdir [UnzipFile $assemblyarchive]} err]} {
	    error "cannot unzip $localname: $err"
	}

	set rootname [file root $localname]
	set cadfilename [file join $tempdir meta-inf $rootname]
	append cadfilename ".cad"

	if {![file exists $cadfilename]} {
	    catch {file delete -force $tempdir}
	    error "no Assembly Descriptor in $url"
	}

	set pkgdir $tempdir
    } else {
	set cadfilename $assemblyarchive
	set pkgdir [file dirname $assemblyarchive]
    }

    set cadlocalname [file tail $cadfilename]
    set status "Reading Assembly Descriptor $cadlocalname ..."
    update

    if {[catch {set file [open $cadfilename]} err]} {
	error "cannot open $cadlocalname: $err"
    }

    set data [read $file]
    close $file

    if {[catch {set doc [::dom::DOMImplementation parse $data]} err]} {
	error "cannot parse contents of $cadlocalname"
    }

    set assembly [::dom::document cget $doc -documentElement]

    if {[::dom::node cget $assembly -nodeName] != "componentassembly"} {
	::dom::DOMImplementation destroy $doc
	error "not an Assembly Descriptor: $cadlocalname"
    }

    #
    # (1) Go through componentfiles and load softpkg files
    #

    set componentfiles [DOMGetElement $assembly componentfiles]

    foreach componentfile [DOMGetAll $componentfiles componentfile] {
	set fileid [::dom::element getAttribute $componentfile id]

	if {![catch {set fileinarchive [DOMGetElement $componentfile fileinarchive]}]} {
	    set softpkgfilename [::dom::element getAttribute $fileinarchive name]
	    if {[catch {set filename [LocateFile $softpkgfilename $pkgdir]} err]} {
		::dom::DOMImplementation destroy $doc
		error "cannot find Softpkg $softpkgfilename: $err"
	    }

	    set localname [file tail $softpkgfilename]
	    set status "Reading Software Package Descriptor $localname ..."
	    update

	    set packnode [FileReadSoftpkg $filename $pkgdir]
	    set idtonode($fileid) $packnode
	} elseif {![catch {set link [DOMGetElement $componentfile link]}]} {
	    set url [::dom::element getAttribute $link href]
	    set status "Reading Software Package Descriptor $url ..."
	    update
	    set packnode [FileReadSoftpkg $url $pkgdir]
	    set idtonode($fileid) $packnode
	}
    }

    #
    # (2) Go through partitioning
    #

    set status "Reading Instantiations ..."
    update

    set partitioning [DOMGetElement $assembly partitioning]

    foreach homeplacement [DOMGetAll $partitioning homeplacement] {
	#
	# make home
	#

	set homenode [FileReadAssemblyLoadHome $homeplacement $pkgdir]

	#
	# traverse componentinstantiations
	#

	foreach ci [DOMGetAll $homeplacement componentinstantiation] {
	    FileReadAssemblyLoadComponent $ci $homenode $pkgdir
	}

	update ;# strategic update
    }

    #
    # (3) Go through connections
    #

    set status "Reading Connections ..."
    update

    set connections [DOMGetElement $assembly connections]
    set connection [::dom::node cget $connections -firstChild]

    while {$connection != ""} {
	switch -- [::dom::node cget $connection -nodeName] {
	    connectinterface -
	    connectevent {
		FileReadAssemblyLoadConnection $connection
	    }
	}
	set connection [::dom::node cget $connection -nextSibling]
	update ;# strategic update
    }

    ::dom::DOMImplementation destroy $doc
}

proc Ass::FileReadAssemblyLoadHome {homeplacement dirname} {
    variable idtonode
    variable homedata
    variable status

    set homeid [::dom::element getAttribute $homeplacement id]
    set componentfileref [DOMGetElement $homeplacement componentfileref]
    set componentfileid [::dom::element getAttribute \
	    $componentfileref idref]

    if {![info exists idtonode($componentfileid)]} {
	error "undefined componentfileref: $componentfileid"
    }

    #
    # Make Home instance
    #

    set packnode $idtonode($componentfileid)
    set homenode [HomeAddHome $packnode $homeid]
    set idtonode($homeid) $homenode

    #
    # Read Home properties
    #

    if {![catch {set homeproperties [DOMGetElement $homeplacement \
	    homeproperties]}]} {
	set fileinarchive [DOMGetElement $homeproperties fileinarchive]
	set propfilename [::dom::element getAttribute \
		$fileinarchive name]

	set localname [file tail $propfilename]
	set status "Reading Property File $localname ..."
	update

	if {[catch {set filename [LocateFile $propfilename $dirname]} err]} {
	    error "cannot find Property File $propfilename: $err"
	}
	
	FileReadProperties $homenode $filename
    }

    #
    # Register with Naming Service?
    #

    if {![catch {set registerwithnaming [DOMGetElement $homeplacement \
	    registerwithnaming]}]} {
	set nsname [::dom::element getAttribute $registerwithnaming name]
	set homedata($homenode,nsreg) 1
	set homedata($homenode,nsname) $nsname
    }

    return $homenode
}

proc Ass::FileReadAssemblyLoadComponent {ci homenode dirname} {
    variable configdata
    variable compdata
    variable idtonode
    variable status

    set compid [::dom::element getAttribute $ci id]
    set compnode [CompAddComponent $homenode $compid]
    set idtonode($compid) $compnode

    #
    # Read Component properties
    #

    if {![catch {set compproperties [DOMGetElement $ci \
	    componentproperties]}]} {
	set fileinarchive [DOMGetElement $compproperties fileinarchive]
	set propfilename [::dom::element getAttribute \
		$fileinarchive name]

	set localname [file tail $propfilename]
	set status "Reading Property File $localname ..."
	update

	if {[catch {set filename [LocateFile $propfilename $dirname]} err]} {
	    error "cannot find Property File $propfilename: $err"
	}
	
	FileReadProperties $compnode $filename
    }

    #
    # Register with Naming Service?
    #

    foreach registercomponent [DOMGetAll $ci registercomponent] {
	update ;# strategic update

	if {[catch {set registerwithnaming [DOMGetElement \
		$registercomponent registerwithnaming]}]} {
	    continue
	}

	set nsname [::dom::element getAttribute $registerwithnaming name]

	if {![catch {set providesidentifier [DOMGetElement \
		$registercomponent providesidentifier]}]} {
	    set textnode [::dom::node cget $providesidentifier -firstChild]
	    set portname [::dom::node cget $textnode -nodeValue]
	    set configdata($compnode,$portname,kind) Provides
	    set configdata($compnode,$portname,nsreg) 1
	    set configdata($compnode,$portname,nsname) $nsname
	} elseif {![catch {set consumesidentifier [DOMGetElement \
		$registercomponent consumesidentifier]}]} {
	    set textnode [::dom::node cget $consumesidentifier -firstChild]
	    set portname [::dom::node cget $textnode -nodeValue]
	    set configdata($compnode,$portname,kind) Consumes
	    set configdata($compnode,$portname,nsreg) 1
	    set configdata($compnode,$portname,nsname) $nsname
	} else {
	    set compdata($compnode,nsreg) 1
	    set compdata($compnode,nsname) $nsname
	}
    }

    return $compnode
}

proc Ass::FileReadAssemblyLoadConnection {connection} {
    variable configdata
    variable idtonode

    set type [::dom::node cget $connection -nodeName]

    if {$type == "connectinterface"} {
	#
	# identify source port
	#

	set sourcekind "Uses"
	set usesport [DOMGetElement $connection usesport]
	set usesidentifier [DOMGetElement $usesport usesidentifier]
	set usesidnode [::dom::node cget $usesidentifier -firstChild]
	set sourceportname [::dom::node cget $usesidnode -nodeValue]

	set cinref [DOMGetElement $usesport componentinstantiationref]
	set sourcecompid [::dom::element getAttribute $cinref idref]
	set sourcecompnode $idtonode($sourcecompid)

	#
	# identify sink
	#

	if {![catch {set providesport [DOMGetElement $connection providesport]}]} {
	    set sinktype "Port"
	    set providesid [DOMGetElement $providesport providesidentifier]
	    set providesidnode [::dom::node cget $providesid -firstChild]
	    set sinkportname [::dom::node cget $providesidnode -nodeValue]

	    set cinref [DOMGetElement $providesport componentinstantiationref]
	    set sinkcompid [::dom::element getAttribute $cinref idref]
	    set sinkcompnode $idtonode($sinkcompid)
	    set sinkdata [list $sinkcompnode $sinkportname]
	    set configdata($sinkcompnode,$sinkportname,kind) "Provides"
	} elseif {![catch {set csuppif [DOMGetElement $connection componentsupportedinterface]}]} {
	    set sinktype "Port"
	    set cinref [DOMGetElement $csuppif componentinstantiationref]
	    set sinkcompid [::dom::element getAttribute $cinref idref]
	    set sinkcompnode $idtonode($sinkcompid)
	    set sinkdata [list $sinkcompnode -1]
	} elseif {![catch {set exif [DOMGetElement $connection existinginterface]}]} {
	    set findby [DOMGetElement $exif findby]
	    if {![catch {set ns [DOMGetElement $findby namingservice]}]} {
		set sinktype "NS"
		set sinkdata [::dom::element getAttribute $ns name]
	    } elseif {![catch {set ref [DOMGetElement $findby stringifiedobjectref]}]} {
		set sinktype "URL"
		set refnode [::dom::node cget $ref -firstChild]
		set sinkdata [::dom::node cget $refnode -nodeValue]
	    } else {
		error "child of existinginterface must by namingservice or stringifiedobjectref"
	    }
	} else {
	    error "connectinterface must have a providesport, componentsupportedinterface or existinginterface node"
	}
    } elseif {$type == "connectevent"} {
	#
	# identify source port
	#

	if {![catch {set emitsport [DOMGetElement $connection emitsport]}]} {
	    set sourcekind "Emits"
	    set emitsid [DOMGetElement $emitsport emitsidentifier]
	    set emitsidnode [::dom::node cget $emitsid -firstChild]
	    set sourceportname [::dom::node cget $emitsidnode -nodeValue]

	    set cinref [DOMGetElement $emitsport componentinstantiationref]
	    set sourcecompid [::dom::element getAttribute $cinref idref]
	    set sourcecompnode $idtonode($sourcecompid)
	} elseif {![catch {set publishesport [DOMGetElement $connection publishesport]}]} {
	    set sourcekind "Publishes"
	    set publishesid [DOMGetElement $publishesport publishesidentifier]
	    set publishesidnode [::dom::node cget $publishesid -firstChild]
	    set sourceportname [::dom::node cget $publishesidnode -nodeValue]

	    set cinref [DOMGetElement $publishesport componentinstantiationref]
	    set sourcecompid [::dom::element getAttribute $cinref idref]
	    set sourcecompnode $idtonode($sourcecompid)
	} else {
	    error "connectevent must have a emitsport or publishesport node"
	}

	#
	# identify sink
	#

	if {![catch {set consumesport [DOMGetElement $connection consumesport]}]} {
	    set sinktype "Port"
	    set consumesid [DOMGetElement $consumesport consumesidentifier]
	    set consumesidnode [::dom::node cget $consumesid -firstChild]
	    set sinkportname [::dom::node cget $consumesidnode -nodeValue]

	    set cinref [DOMGetElement $consumesport componentinstantiationref]
	    set sinkcompid [::dom::element getAttribute $cinref idref]
	    set sinkcompnode $idtonode($sinkcompid)
	    set sinkdata [list $sinkcompnode $sinkportname]
	    set configdata($sinkcompnode,$sinkportname,kind) "Consumes"
	} elseif {![catch {set exif [DOMGetElement $connection existinginterface]}]} {
	    set findby [DOMGetElement $exif findby]
	    if {![catch {set ns [DOMGetElement $findby namingservice]}]} {
		set sinktype "NS"
		set sinkdata [::dom::element getAttribute $ns name]
	    } elseif {![catch {set ref [DOMGetElement $findby stringifiedobjectref]}]} {
		set sinktype "URL"
		set refnode [::dom::node cget $ref -firstChild]
		set sinkdata [::dom::node cget $refnode -nodeValue]
	    } else {
		error "child of existinginterface must by namingservice or stringifiedobjectref"
	    }
	} else {
	    error "connectevent must have a consumesport or existinginterface node"
	}
    } else {
	#
	# neither connectinterface nor connectevent
	#
	return
    }

    if {![info exists configdata($sourcecompnode,$sourceportname,conn)]} {
	set numconn 0
	set configdata($sourcecompnode,$sourceportname,conn) 1
    } else {
	set numconn $configdata($sourcecompnode,$sourceportname,conn)
	incr configdata($sourcecompnode,$sourceportname,conn)
    }

    set configdata($sourcecompnode,$sourceportname,kind) $sourcekind
    set configdata($sourcecompnode,$sourceportname,$numconn,type) $sinktype
    set configdata($sourcecompnode,$sourceportname,$numconn,sink) $sinkdata
}

proc Ass::FileReadSoftpkg {url {dirname ""}} {
    variable nodecounter
    variable packdata

    set softpkgarchive [DownloadFile $url]

    if {[file extension $softpkgarchive] == ".zip"} {
	set softpkgdir [UnzipFile $softpkgarchive]
	set lastidx [string last / $url]
	incr lastidx
	set softpkgurlname [string range $url $lastidx end]
	set softpkgfilename [file rootname $softpkgurlname]
	append softpkgfilename ".csd"
	set softpkgfile [LocateFile $softpkgfilename $softpkgdir]
    } else {
	set softpkgfile $softpkgarchive
	set softpkgdir $dirname
    }

    set file [open $softpkgfile]
    set data [read $file]
    close $file

    set doc [::dom::DOMImplementation parse $data]
    set softpkg [::dom::document cget $doc -documentElement]
    set softpkgname [::dom::element getAttribute $softpkg name]

    set idl [DOMGetElement $softpkg idl]
    set homerepoid [::dom::element getAttribute $idl id]
    if {![catch {set fileinarchive [DOMGetElement $idl fileinarchive]}]} {
        set idlfilename [::dom::element getAttribute $fileinarchive name]
        set idlurl [LocateFile $idlfilename $softpkgdir]
    } elseif {![catch {set link [DOMGetElement $idl link]}]} {
        set idlurl [::dom::element getAttribute $link href]
    } else {
        error "cannot locate IDL file in softpkg $url"
    }

    set implementation [DOMGetElement $softpkg implementation]
    set code [DOMGetElement $implementation code]
    set codetype [::dom::element getAttribute $code type]

    if {![catch {set fileinarchive [DOMGetElement $code fileinarchive]}]} {
        set libfilename [::dom::element getAttribute $fileinarchive name]
        set liburl [LocateFile $libfilename $softpkgdir]
    } elseif {![catch {set link [DOMGetElement $code link]}]} {
        set liburl [::dom::element getAttribute $link href]
    } else {
        error "cannot locate lib file for softpkg $filename"
    }

    if {$codetype == "DLL"} {
	set entrypoint [DOMGetElement $implementation entrypoint]
	set textnode [::dom::node cget $entrypoint -firstChild]
	set entrypt [::dom::node cget $textnode -nodeValue]
    } else {
	set entrypt ""
    }

    #
    # Add Package Node
    #

    set localname [file root [file tail $softpkgarchive]]
    set packnode [PackAddPackage $softpkgname Softpkg $url]

    #
    # Add Node for Home
    #

    if {$codetype == "DLL"} {
	set flatname [string range $entrypt 7 end]
	set homename [string map {_ ::} $flatname]

	set lastidx [string last : $homename]
	incr lastidx
	set text [string range $homename $lastidx end]
    } else {
	set text $softpkgname
    }

    incr nodecounter
    lappend packdata($packnode,children) $nodecounter
    set packdata($nodecounter,nodetype) Home
    set packdata($nodecounter,id) $softpkgname
    set packdata($nodecounter,text) $text
    set packdata($nodecounter,parent) $packnode
    set packdata($nodecounter,codetype) $codetype
    set packdata($nodecounter,whatref) "noref"
    set packdata($nodecounter,pkgurl) $url
    set packdata($nodecounter,idlurl) $idlurl
    set packdata($nodecounter,liburl) $liburl
    set packdata($nodecounter,repoid) $homerepoid
    set packdata($nodecounter,entrypt) $entrypt
    set packdata($nodecounter,idlintrospected) 0
    set packdata($nodecounter,idllocalfile) ""

    #
    # clear up doc
    #

    ::dom::DOMImplementation destroy $doc
    return $nodecounter
}

proc Ass::FileReadProperties {parentnode filename} {
    variable configdata

    set file [open $filename]
    set data [read $file]
    close $file

    set doc [::dom::DOMImplementation parse $data]
    set properties [::dom::document cget $doc -documentElement]

    set property [::dom::node cget $properties -firstChild]

    while {$property != ""} {
	switch -- [::dom::node cget $property -nodeName] {
	    simple -
	    sequence -
	    struct -
	    valuetype {
		set attrname [::dom::element getAttribute $property name]
		set configdata($parentnode,$attrname,kind) "Attribute"
		set configdata($parentnode,$attrname,set) 1
		FileReadPropertyItem $property $parentnode \
			$attrname $attrname
	    }
	}

	set property [::dom::node cget $property -nextSibling]
    }

    ::dom::DOMImplementation destroy $doc
}

proc Ass::FileReadPropertyItem {property id attrname name} {
    variable configdata

    set kind [::dom::node cget $property -nodeName]

    if {$kind == "simple"} {
	set type [::dom::element getAttribute $property type]
	set valuenode [DOMGetElement $property value]
	set valuetext [::dom::node cget $valuenode -firstChild]
	set value [::dom::node cget $valuetext -nodeValue]

	set configdata($id,$attrname,type,$name) $type
	set configdata($id,$attrname,value,$name) $value
    } elseif {$kind == "sequence"} {
	set element [::dom::node cget $property -firstChild]
	set length 0

	while {$element != ""} {
	    switch -- [::dom::node cget $element -nodeName] {
		simple -
		sequence -
		struct -
		valuetype {
		    set subname "$name\[$length\]"
		    FileReadPropertyItem $element $id $attrname $subname
		    incr length
		}
	    }
	    set element [::dom::node cget $element -nextSibling]
	}

	set configdata($id,$attrname,type,$name) sequence
	set configdata($id,$attrname,value,$name) $length

    } elseif {$kind == "struct" || $kind == "valuetype"} {
	set element [::dom::node cget $property -firstChild]
	set length 0

	set configdata($id,$attrname,type,$name) $kind
	set configdata($id,$attrname,members,$name) [list]

	while {$element != ""} {
	    switch -- [::dom::node cget $element -nodeName] {
		simple -
		sequence -
		struct -
		valuetype {
		    set membername [::dom::element getAttribute $element name]
		    lappend configdata($id,$attrname,members,$name) $membername
		    set subname "$name.$membername"
		    FileReadPropertyItem $element $id $attrname $subname
		}
	    }
	    set element [::dom::node cget $element -nextSibling]
	}
    }
}

#
# ======================================================================
# Deploy Assembly
# ======================================================================
#

proc Ass::DeployAssembly {ccmd collocate {callcc 1}} {
    variable configdata
    variable packdata
    variable homedata
    variable compdata
    variable compserverrefs
    variable containerrefs
    variable uploadedpacks
    variable comprefs
    variable homerefs
    variable status
    variable setup

    #
    # init
    #

    set compserverrefs [list]
    set uploadedpacks [list]
    catch {unset containerrefs}
    catch {unset comprefs}
    catch {unset homerefs}

    #
    # (1) Upload Softpkg files
    #

    set deployablehomecount 0

    foreach homenode $homedata(allhomes) {
	#
	# Upload Package to Component Installation
	#

	set packhomenode $homedata($homenode,package)
	set homeentrypt $packdata($packhomenode,entrypt)
	set packliburl $packdata($packhomenode,liburl)
	set codetype $packdata($packhomenode,codetype)

	set lastidx [string last / $packliburl]
	incr lastidx
	set implid [string range $packliburl $lastidx end]

	if {![info exists uploaded($implid)]} {
	    set status "Uploading $implid ..."
	    update ;# strategic update

	    if {[file exists $packliburl]} {
		::fpsfhttpd::serve $setup(port) $packliburl $implid
		set liburl "http://"
		if {$setup(host) == "<auto>" || $setup(host) == ""} {
		    append liburl [info hostname]
		} else {
		    append liburl $setup(host)
		}
		append liburl ":" $setup(port)
		append liburl "/" $implid
	    } else {
		set liburl $packliburl
	    }

	    corba::try {
		$ccmd install $implid $liburl
	    } catch {... err} {
		catch {
		    ::fpsfhttpd::stop
		    ::fpsfhttpd::wait
		}
		error "installing $liburl as $implid failed: $err"
	    }

	    if {[file exists $packliburl]} {
		::fpsfhttpd::stop
		::fpsfhttpd::wait
	    }

	    lappend uploadedpacks $implid
	    set uploaded($implid) 1
	}

	set impluuid($homenode) $implid

	if {$codetype == "DLL"} {
	    incr deployablehomecount
	}
    }

    #
    # (2) Make Home instances and configure them
    #

    if {$collocate && $deployablehomecount} {
	set componentserver [$ccmd create_component_server {}]
	lappend compserverrefs $componentserver
    }

    foreach homenode $homedata(allhomes) {
	set homeid $homedata($homenode,id)
	set packhomenode $homedata($homenode,package)
	set codetype $packdata($packhomenode,codetype)
	set entrypt $packdata($packhomenode,entrypt)

	set status "Deploying Home $homeid ..."
	update ;# strategic update

	if {$codetype == "DLL"} {
	    if {!$collocate} {
		set componentserver [$ccmd create_component_server {}]
		lappend compserverrefs $componentserver
	    }

	    set container [$componentserver create_container {}]
	    set containerref($homenode) $container
	    set homeref($homenode) [$container install_home \
		    $impluuid($homenode) $entrypt {}]
	} else {
	    set reffilename [format "%s.ref" $impluuid($homenode)]
	    set homerepoid $packdata($packhomenode,repoid)
	    set homeref($homenode) [$ccmd exec $impluuid($homenode) \
		    [list --ior $reffilename $homerepoid] \
		    $reffilename]
	}

	if {[llength [array names configdata $homenode,*,set]] > 0} {
	    set status "Configuring Home $homeid ..."
	}

	DeployConfigure homedata $homenode $homeref($homenode)
    }

    #
    # (3) Make Component instances
    #

    foreach compnode $compdata(allcomponents) {
	set compid $compdata($compnode,id)
	set homenode $compdata($compnode,home)
	set status "Deploying Component $compid ..."
	update ;# strategic update

	set comprefs($compnode) [corba::dii $homeref($homenode) {Object create {}}]

	if {[llength [array names configdata $compnode,*,set]] > 0} {
	    set status "Configuring Component $compid ..."
	    update ;# strategic update
	}

	DeployConfigure compdata $compnode $comprefs($compnode)
    }

    #
    # (4) Configure Component Ports
    #     - Naming Service registration
    #     - connections
    #

    foreach compnode $compdata(allcomponents) {
	DeployConfigureComponentPorts $compnode $comprefs($compnode)
    }

    #
    # (5) Call configuration_complete
    #

    if {$callcc} {
	foreach compnode $compdata(allcomponents) {
	    set compid $compdata($compnode,id)
	    set status "Calling configuration_complete for Component $compid ..."
	    corba::dii $comprefs($compnode) {void configuration_complete {}}
	    update ;# strategic update
	}
    }
}

proc Ass::DeployConfigure {homeorcompdata parentnode ref} {
    variable configdata
    variable status
    upvar $homeorcompdata data

    #
    # register in the Naming Service?
    #

    set id $data($parentnode,id)

    if {[info exists data($parentnode,nsreg)] && \
	    [info exists data($parentnode,nsname)] && \
	    $data($parentnode,nsreg) && \
	    $data($parentnode,nsname) != ""} {
	set status "Registering $id as $data($parentnode,nsname)"
	DeployMakeNSBinding $data($parentnode,nsname) $ref
    }

    #
    # browse attributes
    #

    foreach arrname [array names configdata $parentnode,*,set] {
	if {[llength [split $arrname ,]] != 3} {
	    continue
	}

	set attrname [lindex [split $arrname ,] 1]

	if {$configdata($parentnode,$attrname,kind) != "Attribute"} {
	    continue
	}

	set any [DeployCollectProperty $parentnode $attrname $attrname]
	set attrtype [lindex $any 0]
	set attrvalue [lindex $any 1]

	corba::try {
	    corba::dii $ref [list void _set_$attrname \
		    [list [list in $attrtype]]] $attrvalue
	} catch {... err} {
	    global errorInfo
	    error "cannot set attribute $attrname of $id: $err" $errorInfo
	}

	update ;# strategic update
    }
}

proc Ass::DeployConfigureComponentPorts {compnode compref} {
    variable configdata
    variable compdata
    variable status

    set id $compdata($compnode,id)

    set status "Connecting Ports for $id ..."

    foreach arrname [array names configdata $compnode,*,kind] {
	if {[llength [split $arrname ,]] != 3} {
	    continue
	}

	set portname [lindex [split $arrname ,] 1]
	set portkind $configdata($compnode,$portname,kind)

	update ;# strategic update

	#
	# Register facets and event sources with Naming Service
	#

	if {$portkind == "Provides" || $portkind == "Consumes"} {
	    if {[info exists configdata($compnode,$portname,nsreg)] && \
		    [info exists configdata($compnode,$portname,nsname)] && \
		    $configdata($compnode,$portname,nsreg) && \
		    $configdata($compnode,$portname,nsname) != ""} {
		#
		# Acquire reference
		#

		switch -- $portkind {
		    Provides {
			set ref [corba::dii $compref {Object provide_facet {{in string}}} $portname]
		    }
		    Consumes {
			set ref [corba::dii $compref {Object get_consumer {{in string}}} $portname]
		    }
		}

		#
		# Register in NS
		#

		set status "Registering port $portname of $id as $configdata($compnode,$portname,nsname)"
		DeployMakeNSBinding $configdata($compnode,$portname,nsname) $ref
	    }
	}

	#
	# For receptacles and event sources, make their connections
	#

	if {($portkind == "Uses" || $portkind == "Emits" || \
		$portkind == "Publishes") && \
		[info exists configdata($compnode,$portname,conn)]} {
	    DeployConfigureConnection $compnode $portname $compref
	}
    }
}

proc Ass::DeployConfigureConnection {compnode portname compref} {
    variable configdata
    variable compdata
    variable comprefs
    variable status

    set compid $compdata($compnode,id)
    set portkind $configdata($compnode,$portname,kind)
    set numconns $configdata($compnode,$portname,conn)

    for {set i 0} {$i < $numconns} {incr i} {
	if {![info exists configdata($compnode,$portname,$i,type)]} {
	    continue
	}

	set sinktype $configdata($compnode,$portname,$i,type)
	set sinkdata $configdata($compnode,$portname,$i,sink)

	if {$sinktype == "none"} {
	    continue
	}

	#
	# Identify sink reference
	#

	switch -- $sinktype {
	    URL {
		corba::try {
		    set sinkref [corba::string_to_object $sourcedata]
		} catch {... err} {
		    global errorInfo
		    error "cannot read URL for conn # $i of port $portname of $compid: $err" $errorInfo
		}
	    }
	    NS {
		corba::try {
		    set ns [corba::resolve_initial_references NameService]
		    set sinkref [corba::dii $ns {Object resolve_str {{in string}}} $sourcedata]
		} catch {... err} {
		    global errorInfo
		    error "cannot find $sourcedata in NamingService for port $portname of $compid: $err" $errorInfo
		}
	    }
	    Port {
		set sinkcompnode [lindex $sinkdata 0]
		set sinkportname [lindex $sinkdata 1]
		set sinkportkind $configdata($sinkcompnode,$sinkportname,kind)
		set sinkcompid $compdata($sinkcompnode,id)

		if {$sinkportname == -1} {
		    set sinkref $comprefs($sinkcompnode)
		} else {
		    set sinkcompref $comprefs($sinkcompnode)

		    switch -- $sinkportkind {
			Provides {
			    set sinkref [corba::dii $sinkcompref {Object provide_facet {{in string}}} $sinkportname]
			}
			Consumes {
			    set sinkref [corba::dii $sinkcompref {Object get_consumer {{in string}}} $sinkportname]
			}
			default {
			    error "oops: $compnode $portname $i $sinkportkind"
			}
		    }
		}
	    }
	}

	#
	# make connection
	#

	switch -- $portkind {
	    Uses {
		corba::dii $compref {{valuetype {} {public CookieValue {sequence octet}}} connect {{in string} {in Object}}} $portname $sinkref
	    }
	    Emits {
		corba::dii $compref {void connect_consumer {{in string} {in Object}}} $portname $sinkref
	    }
	    Publishes {
		corba::dii $compref {{valuetype {} {public CookieValue {sequence octet}}} subscribe {{in string} {in Object}}} $portname $sinkref
	    }
	    default {
		error "oops: $compnode $portname $i $portkind"
	    }
	}
    }
}

proc Ass::DeployCollectProperty {id attrname name} {
    variable configdata

    set kind $configdata($id,$attrname,type,$name)

    switch -- $kind {
	boolean -
	short -
	long -
	float -
	double -
	char -
	octet -
	wchar -
	string -
	wstring {
	    set type $configdata($id,$attrname,type,$name)
	    set value $configdata($id,$attrname,value,$name)
	    return [list $type $value]
	}
	ushort {
	    set type "unsigned short"
	    set value $configdata($id,$attrname,value,$name)
	    return [list $type $value]
	}
	ulong {
	    set type "unsigned long"
	    set value $configdata($id,$attrname,value,$name)
	    return [list $type $value]
	}
	longlong {
	    set type "long long"
	    set value $configdata($id,$attrname,value,$name)
	    return [list $type $value]
	}
	ulonglong {
	    set type "unsigned long long"
	    set value $configdata($id,$attrname,value,$name)
	    return [list $type $value]
	}
	longdouble {
	    set type "long double"
	    set value $configdata($id,$attrname,value,$name)
	    return [list $type $value]
	}
    }
    
    if {$kind == "sequence"} {
	set length $configdata($id,$attrname,value,$name)
	set subtype "dummy"
	set elements [list]

	for {set i 0} {$i < $length} {incr i} {
	    set subname "$name\[$i\]"
	    set subany [DeployCollectProperty $id $attrname $subname]
	    set subtype [lindex $subany 0]
	    set subvalue [lindex $subany 1]
	    lappend elements $subvalue
	}

	return [list [list sequence $subtype] $elements]
    }
    elseif {$kind == "struct" || $kind == "valuetype"} {
	set membertypes [list]
	set membervalues [list]

	foreach membername $configdata($id,$attrname,members,$name) {
	    set subname "$name.$membername"
	    set subany [DeployCollectProperty $id $attrname $subname]
	    set subtype [lindex $subany 0]
	    set subvalue [lindex $subany 1]
	    lappend membertypes $subname $subtype
	    lappend membervalues $subname $subvalue
	}

	return [list [list $kind "" $membertypes] $membervalues]
    }
    else {
	error "oops"
    }
}

proc Ass::DeployMakeNSBinding {nsname obj} {
    corba::try {
	set ns [corba::resolve_initial_references NameService]
	set name [corba::dii $ns {{sequence {struct {} {id string kind string}}} to_name {{in string}}} $nsname]

	#
	# make intermediate contexts
	#

	set namelen [llength $name]
	set contexts [lrange $name 0 [expr {$namelen-1}]]
	set curcontext $ns

	foreach context $contexts {
	    corba::try {
		set newcontext [corba::dii $curcontext {Object resolve {{in {sequence {struct {} {id string kind string}}}}}} [list $context]]
	    } catch {...} {
		set newcontext [corba::dii $curcontext {Object bind_new_context {{in {sequence {struct {} {id string kind string}}}}}} [list $context]]
	    }
	    set curcontext $newcontext
	}

	corba::dii $ns {void rebind {{in {sequence {struct {} {id string kind string}}}} {in Object}}} $name $obj
    } catch {... err} {
	global errorInfo
	error "cannot register $nsname in Naming Service: $err" $errorInfo
    }
}

#
# ======================================================================
# Tear Down Assembly
# ======================================================================
#

proc Ass::TearDownAssembly {ccmd} {
    variable compserverrefs
    variable containerrefs
    variable uploadedpacks
    variable comprefs
    variable homerefs
    variable status

    #
    # (1) remove all components
    #

    if {[info exists comprefs]} {
	set status "Removing components ..."
	update ;# strategic update

	foreach compnode [array names comprefs] {
	    corba::dii $comprefs($compnode) {void remove {}}
	}
    }

    #
    # (2) remove all homes
    #

    if {[info exists homerefs]} {
	set status "Removing homes ..."
	update ;# strategic update

	foreach homenode [array names homerefs] {
	    $containerrefs($homenode) remove_home $homerefs($homenode)
	}
    }

    #
    # (3) remove all containers
    #

    if {[info exists homerefs]} {
	set status "Removing containers ..."
	update ;# strategic update

	foreach homenode [array names homerefs] {
	    $containerrefs($homenode) remove
	}
    }

    #
    # (4) remove all component servers
    #

    set status "Removing Component Servers ..."
    update ;# strategic update

    foreach compserver $compserverrefs {
	$compserver remove
    }

    #
    # (5) remove component installations
    #

    set status "Removing Implementations ..."
    update ;# strategic update

    foreach pack $uploadedpacks {
	$ccmd remove $pack
    }
}
