#
# ======================================================================
# This file is part of Combat/Tcl, a Tcl CORBA Object Request Broker
#
# Please visit the Combat Web site at http://www.fpx.de/Combat/ for
# more information.
#
# Copyright (c) Frank Pilhofer, combat@fpx.de
#
# ======================================================================
#
# CVS Version Tag: $Id: iop.tcl,v 1.5 2002/02/26 10:50:59 fp Exp $
#
# ----------------------------------------------------------------------
# IOP Module
# ----------------------------------------------------------------------
#

namespace eval Combat {
    namespace eval IOP {
	class TaggedComponent {
	    public variable tag
	    public variable component_data

	    public method marshal {buffer} {
		$buffer ulong $tag
		$buffer ulong [string length $component_data]
		$buffer octets $component_data
	    }
	}

	proc DemarshalTaggedComponent {buffer} {
	    set tag [$buffer ulong]

	    switch -- $tag {
		1 { # TAG_CODE_SETS
		    set res [::Combat::CONV_FRAME::DemarshalCodeSetInfo $buffer]
		}
		default {
		    set length [$buffer ulong]
		    set component_data [$buffer octets $length]
		    set res [namespace current]::[TaggedComponent #auto]
		    $res configure -tag $tag -component_data $component_data
		}
	    }

	    return $res
	}

	class ServiceContext {
	    public variable context_id
	    public variable context_data

	    public method marshal {buffer} {
		$buffer ulong $context_id
		$buffer ulong [string length $context_data]
		$buffer octets $context_data
	    }
	}

	proc DemarshalServiceContext {buffer} {
	    set context_id [$buffer ulong]

	    switch -- $context_id {
		1 { # CodeSets
		    set res [::Combat::CONV_FRAME::DemarshalCodeSetContext $buffer]
		}
		default {
		    set length [$buffer ulong]
		    set context_data [$buffer octets $length]
		    set res [namespace current]::[ServiceContext #auto]
		    $res configure -context_id $context_id \
			    -context_data $context_data
		}
	    }

	    return $res
	}

	class TaggedProfile {
	    public variable tag
	    public variable profile_data

	    public method connect {} {
		error "not reachable"
	    }

	    public method marshal {buffer} {
		$buffer ulong $tag
		$buffer ulong [string length $profile_data]
		$buffer octets $profile_data
	    }
	}

	class MultipleComponentProfile {
	    inherit ::Combat::IOP::TaggedProfile
	    public variable components

	    constructor {} {
		set components [list]
	    }

	    destructor {
		foreach component $components {
		    delete object $component
		}
	    }

	    public method marshal {buffer} {
		$buffer ulong 1		;# TAG_MULTIPLE_COMPONENTS
		$buffer begin_encaps
		$buffer ulong [llength $components]
		foreach component $components {
		    $component marshal $buffer
		}
		$buffer end_encaps
	    }
	}

	proc DemarshalMultipleComponentProfile {buffer} {
	    $buffer begin_encaps
	    set components [list]
	    set count [$buffer ulong]
	    for {set i 0} {$i < $count} {incr i} {
		lappend components [::Combat::IOP::DemarshalTaggedComponent $buffer]
	    }
	    $buffer end_encaps

	    set res [namespace current]::[::Combat::IOP::MultipleComponentProfile #auto]
	    $res configure -tag 1 -components $components
	    return $res
	}

	proc DemarshalTaggedProfile {buffer} {
	    set tag [$buffer ulong]

	    switch -- $tag {
		0 { # TAG_INTERNET_IOP
		    set res [::Combat::IIOP::DemarshalIIOPProfile $buffer]
		}
		1 { # TAG_MULTIPLE_COMPONENTS
		    set res [::Combat::IOP::DemarshalMultipleComponentProfile $buffer]
		}
		default {
		    set length [$buffer ulong]
		    set profile_data [$buffer octets $length]
		    set res [namespace current]::[TaggedProfile #auto]
		    $res configure -tag $tag -profile_data $profile_data
		}
	    }

	    return $res
	}

	class IOR {
	    public variable type_id
	    public variable profiles

	    constructor {} {
		set type_id ""
		set profiles [list]
	    }

	    destructor {
		foreach profile $profiles {
		    delete object $profile
		}
	    }

	    public method marshal {buffer} {
		$buffer string $type_id
		$buffer ulong [llength $profiles]
		foreach profile $profiles {
		    $profile marshal $buffer
		}
	    }

	    public method stringify {} {
		set buffer [namespace current]::[::Combat::CDR::WriteBuffer #auto]
		$buffer boolean [$buffer cget -byteorder]
		marshal $buffer
		set data [$buffer get_data]
		binary scan $data H* hex
		delete object $buffer
		return "IOR:$hex"
	    }

	    #
	    # Find Codeset info in a MultipleComponents profile.
	    #

	    public method getCodesetInfo {} {
		foreach profile $profiles {
		    # test for TAG_MULTIPLE_COMPONENTS
		    if {[$profile cget -tag] == 1} {
			foreach component [$profile cget -components] {
			    # test for TAG_CODE_SETS
			    if {[$component cget -tag] == 1} {
				return $component
			    }
			}
		    }
		}
		return ""
	    }
	}

	proc DemarshalIOR {buffer} {
	    set type_id [$buffer string]
	    set profiles [list]
	    set numofprofiles [$buffer ulong]
	    for {set i 0} {$i < $numofprofiles} {incr i} {
		lappend profiles [::Combat::IOP::DemarshalTaggedProfile $buffer]
	    }
	    set ior [namespace current]::[::Combat::IOP::IOR #auto]
	    $ior configure -type_id $type_id -profiles $profiles
	    return $ior
	}

	proc DestringifyIOR {ior} {
	    if {[string range $ior 0 3] != "IOR:"} {
		error "not a stringified IOR: $ior"
	    }

	    set hex [string range $ior 4 end]
	    set data [binary format H* $hex]
	    set buffer [namespace current]::[::Combat::CDR::ReadBuffer #auto $data]
	    set byteorder [$buffer boolean]
	    $buffer configure -byteorder $byteorder
	    set res [DemarshalIOR $buffer]
	    delete object $buffer
	    return $res
	}
    }
}

