#
# ======================================================================
# This file is part of Combat/Tcl, a Tcl CORBA Object Request Broker
#
# Please visit the Combat Web site at http://www.fpx.de/Combat/ for
# more information.
#
# Copyright (c) Frank Pilhofer, combat@fpx.de
#
# ======================================================================
#
# CVS Version Tag: $Id: orb.tcl,v 1.10 2002/03/12 10:18:18 fp Exp $
#
# ----------------------------------------------------------------------
# The ORB
# ----------------------------------------------------------------------
#

namespace eval Combat {
    namespace eval CORBA {
	variable ORB_conn_cond

	class ORB {
	    private common requests
	    private common callbacks
	    private common asyncid 0
	    private common asyncids_fwd
	    private common asyncids_rev
	    private common fake_asyncs
	    private common objprocid 0
	    public common handles_fwd
	    public common handles_rev
	    public common initial_pseudo_references
	    public common initial_references
	    public common default_init_ref ""
	    public common iiop_serverports
	    public common iiop_servers
	    public common ior_template

	    #
	    # send_request and receive_reply are low-level functions
	    # that process a single GIOP request/response.
	    #
	    # They are encapsulated by the higher-level functions
	    # invoke_sync, invoke_async, get_async_reply and
	    # poll_async_reply, which transparently perform a retry
	    # upon LOCATION_FORWARD etc.
	    #

	    private proc send_request {response object operation rtype \
		    plevel params values {exceptions {}} {callback {}}} {
		incr plevel
		set conn [$object connect]

		if {$callback != "" && $response} {
		    set id [$conn send_request $response $object \
			    $operation $plevel $params $values {} \
			    "::Combat::CORBA::ORB::callback"]
		} else {
		    set id [$conn send_request $response $object \
			    $operation $plevel $params $values]
		}

		if {$response} {
		    $conn ref
		    set requests($id) [list $object $operation $conn $rtype \
			    $params $values $exceptions $callback]
		} else {
		    set requests($id) [list $object {} {} {} \
			    {} {} {} $callback]
		}

		#
		# trigger callback if !response_expected
		#

		if {!$response && $callback != ""} {
		    after 0 "::Combat::CORBA::ORB::callback $id 0"
		}

		return $id
	    }

	    private proc receive_reply {id plevel} {
		if {![info exists requests($id)]} {
		    ::corba::throw [list IDL:omg.org/CORBA/BAD_PARAM:1.0 \
			    [list minor 0 completion_status COMPLETED_MAYBE]]
		}

		set data $requests($id)
		unset requests($id)

		set object    [lindex $data 0]
		set operation [lindex $data 1]
		set conn      [lindex $data 2]
		set rtype     [lindex $data 3]
		set params    [lindex $data 4]
		set values    [lindex $data 5]
		set exceptions [lindex $data 6]
		set callback  [lindex $data 7]

		if {$conn == ""} {
		    # oneway op
		    return [list 0 ""]
		}

		incr plevel
		set res [$conn receive_reply $id $rtype $plevel \
			$params $values $exceptions]

		$conn deref

		set status [lindex $res 0]
		switch -- $status {
		    0 -
		    1 -
		    2 {
			# After a successful method invocation, we set a
			# flag that allows the object to reuse its entire
			# set of profiles after disconnecting
			$object configure -reset_profile 1
		    }
		    3 { # LOCATION_FORWARD
			#
			# A LOCATION_FORWARD is also indicated, without a
			# new IOR, upon receiving a CloseConnection message.
			#
			if {[llength $res] == 2} {
			    $object forward [lindex $res 1]
			} else {
			    $object next_profile
			}
		    }
		    4 { # LOCATION_FORWARD_PERM
			$object forward [lindex $res 1] 1
		    }
		    5 { # NEEDS_ADDRESSING_MODE
			$object configure -AddressingDisposition [lindex $res 1]
		    }
		}

		#
		# Resend request
		#

		if {$status == 3 || $status == 4 || $status == 5} {
		    if {$callback != ""} {
			set newid [send_request 1 $object $operation $rtype \
				$plevel $params $values $exceptions \
				$callback]
		    } else {
			set newid [send_request 1 $object $operation $rtype \
				$plevel $params $values $exceptions]
		    }

		    return [list $status $newid]
		}

		return $res
	    }

	    #
	    # Synchronous invocation of requests
	    #

	    public proc invoke_sync {response object operation rtype \
		    plevel params values {exceptions {}}} {
		incr plevel
		set id [send_request $response $object $operation $rtype \
			$plevel $params $values $exceptions]
		#
		# oneway ?
		#

		if {!$response} {
		    return
		}

		#
		# wait for reply
		#

		set res [receive_reply $id $plevel]

		#
		# Transparently try again in case of LOCATION_FORWARD,
		# LOCATION_FORWARD_PERM or NEEDS_ADDRESSING_MODE
		#

		set status [lindex $res 0]
		while {$status == 3 || $status == 4 || $status == 5} {
		    set newid [lindex $res 1]
		    set res [receive_reply $newid $plevel]
		    set status [lindex $res 0]
		}
		
		#
		# Evaluate response
		#

		switch -- $status {
		    0 { # NO_EXCEPTION
			return [lindex $res 1]
		    }
		    1 { # USER_EXCEPTION
			error [lindex $res 1]
		    }
		    2 { # SYSTEM_EXCEPTION
			error [lindex $res 1]
		    }
		    default {
			::corba::throw [list IDL:omg.org/CORBA/INTERNAL:1.0 \
				[list minor 0 completion_status COMPLETED_NO \
				debug "status is $status"]]
		    }
		}

		::corba::throw [list IDL:omg.org/CORBA/INTERNAL:1.0 \
			[list minor 0 completion_status COMPLETED_NO]]
	    }

	    #
	    # Asynchronous invocation
	    #

	    public proc invoke_async {response object operation rtype \
		    plevel params values {exceptions {}} {callback {}}} {
		incr plevel
		set id [send_request $response $object $operation $rtype \
			$plevel $params $values $exceptions $callback]

		incr asyncid
		set asyncids_fwd($asyncid) $id
		set asyncids_rev($id) $asyncid
		set callbacks($asyncid) $callback
		return $asyncid
	    }

	    public proc fake_async {status result {callback {}}} {
		incr asyncid
		set asyncids_fwd($asyncid) -1
		set fake_asyncs($asyncid) [list $status $result $callback]
		if {$callback != ""} {
		    after 0 "::Combat::CORBA::ORB::callback fake:$asyncid $status"
		}
		return $asyncid
	    }

	    public proc get_async_reply {aid plevel} {
		if {![info exists asyncids_fwd($aid)]} {
		    ::corba::throw [list IDL:omg.org/CORBA/BAD_PARAM:1.0 \
			    [list minor 0 completion_status COMPLETED_MAYBE]]
		}

		set id $asyncids_fwd($aid)

		if {$id == -1} {
		    set res $fake_asyncs($aid)
		    unset asyncids_fwd($aid)
		    unset fake_asyncs($aid)
		    set status [lindex $res 0]

		    switch -- $status {
			0 { # NO_EXCEPTION
			    return [lindex $res 1]
			}
			1 { # USER_EXCEPTION
			    error [lindex $res 1]
			}
			2 { # SYSTEM_EXCEPTION
			    error [lindex $res 1]
			}
		    }

		    error "oops"
		}

		unset asyncids_fwd($aid)
		unset asyncids_rev($id)

		incr plevel
		set res [receive_reply $id $plevel]

		#
		# Transparently try again in case of LOCATION_FORWARD,
		# LOCATION_FORWARD_PERM or NEEDS_ADDRESSING_MODE
		#

		set status [lindex $res 0]
		while {$status == 3 || $status == 4 || $status == 5} {
		    set newid [lindex $res 1]
		    set res [receive_reply $newid $plevel]
		    set status [lindex $res 0]
		}

		#
		# Evaluate response
		#

		switch -- $status {
		    0 { # NO_EXCEPTION
			return [lindex $res 1]
		    }
		    1 { # USER_EXCEPTION
			error [lindex $res 1]
		    }
		    2 { # SYSTEM_EXCEPTION
			error [lindex $res 1]
		    }
		    default {
			::corba::throw [list IDL:omg.org/CORBA/INTERNAL:1.0 \
				[list minor 0 completion_status COMPLETED_NO \
				debug "status is $status"]]
		    }
		}

		::corba::throw [list IDL:omg.org/CORBA/INTERNAL:1.0 \
			[list minor 0 completion_status COMPLETED_NO]]
	    }

	    public proc poll_async_reply {aid} {
		if {![info exists asyncids_fwd($aid)]} {
		    ::corba::throw [list IDL:omg.org/CORBA/BAD_PARAM:1.0 \
			    [list minor 0 completion_status COMPLETED_MAYBE]]
		}

		set id $asyncids_fwd($aid)

		if {$id == -1} {
		    return 1
		}

		if {![info exists requests($id)]} {
		    ::corba::throw [list IDL:omg.org/CORBA/BAD_PARAM:1.0 \
			    [list minor 0 completion_status COMPLETED_MAYBE]]
		}

		set conn [lindex $requests($id) 2]
		set status [$conn poll_reply $id]

		#
		# resend request in case of LOCATION_FORWARD etc.
		#

		switch -- $status {
		    0 -
		    1 -
		    2 {
			return 1
		    }
		    3 -
		    4 -
		    5 {
			unset asyncids_rev($id)
			set res [receive_reply $id 0]
			set newid [lindex $res 1]
			set asyncids_fwd($aid) $newid
			set asyncids_rev($newid) $aid
			return 0
		    }
		    -1 {
			return 0
		    }
		    default {
			error "oops"
		    }
		}

		return 0
	    }

	    #
	    # Incoming invocation from GIOP
	    #

	    public proc invoke {response_expected \
		    conn request_id object_key \
		    operation decoder contexts} {

		#
		# gotta hand this over to an object adapter
		#

		::Combat::PortableServer::RootPOA::invoke $response_expected \
			$conn $request_id $object_key \
			$operation $decoder $contexts
	    }

	    public proc locate {conn request_id object_key} {
		#
		# gotta hand this over to an object adapter
		#

		::Combat::PortableServer::RootPOA::locate \
			$conn $request_id $object_key
	    }

	    #
	    # Callback from GIOP connection
	    #

	    public proc callback {id status} {
		#
		# Transparently try again in case of LOCATION_FORWARD,
		# LOCATION_FORWARD_PERM or NEEDS_ADDRESSING_MODE. This
		# retry is performed by receive_reply.
		#

		if {[string range $id 0 3] == "fake"} {
		    set aid [lindex [split $id :] 1]
		} else {
		    set aid $asyncids_rev($id)
		}

		if {$status == 3 || $status == 4 || $status == 5} {
		    unset asyncids_rev($id)
		    set res [receive_reply $id 0]
		    set newid [lindex $res 1]
		    set asyncids_fwd($aid) $newid
		    set asyncids_rev($newid) $aid
		    return
		}

		#
		# Else perform the upcall
		#

		set callback $callbacks($aid)
		unset callbacks($aid)
		uplevel #0 "$callback $aid"
	    }

	    #
	    # Handle Handling
	    #

	    public proc MakeObjProc {obj} {
		incr objprocid
		set handle "_combat_obj_$objprocid"
		uplevel #0 proc $handle {args} \{ \
			eval ::Combat::CORBA::ORB::object_invoke $obj \$args \
			\}
		set handles_fwd($handle) $obj
		set handles_rev($obj) $handle
		return $handle
	    }

	    public proc GetObjFromHandle {handle} {
		if {$handle == "0"} {
		    return 0
		}
		if {![info exists handles_fwd($handle)]} {
		    ::corba::throw [list IDL:omg.org/CORBA/BAD_PARAM:1.0 \
			    [list minor 0 completion_status COMPLETED_NO]]
		}
		return $handles_fwd($handle)
	    }

	    public proc GetHandleFromObj {obj} {
		if {$handle == "0"} {
		    return 0
		}
		if {![info exists handles_rev($obj)]} {
		    ::corba::throw [list IDL:omg.org/CORBA/BAD_PARAM:1.0 \
			    [list minor 0 completion_status COMPLETED_NO]]
		}
		return $handles_rev($obj)
	    }

	    public proc ReleaseHandle {handle} {
		if {$handle == "0"} {
		    return
		}
		set obj [GetObjFromHandle $handle]
		unset handles_fwd($handle)
		unset handles_rev($obj)
		delete object $obj
		uplevel #0 rename $handle \"\"
	    }

	    #
	    # Method invocation
	    #

	    public proc object_invoke {obj args} {
		set idx 0
		set async 0
		set callback ""
		set response 1

		if {[lindex $args $idx] == "-async"} {
		    set async 1
		    incr idx
		} elseif {[lindex $args $idx] == "-callback"} {
		    incr idx
		    set async 1
		    set callback [lindex $args $idx]
		    incr idx
		}

		set operation [lindex $args $idx]
		incr idx
		set values [lrange $args $idx end]

		if {$operation == ""} {
		    error "usage: $handles_rev($obj) op ?parameters ...?"
		}

		if {$operation == "_is_a" || \
			$operation == "_get_interface" || \
			$operation == "_non_existent" || \
			$operation == "_duplicate"} {
		    switch -- $operation {
			_is_a {
			    if {[llength $values] != 1} {
				error "usage: $handles_rev($obj) _is_a repoid"
			    }
			    if {!$async} {
				return [$obj _is_a [lindex $values 0]]
			    }
			    set theop "_is_a"
			    set rtype "boolean"
			    set params {{in string}}
			}
			_get_interface {
			    if {[llength $values] != 0} {
				error "usage: $handles_rev($obj) _get_interface"
			    }
			    if {!$async} {
				return [$obj _get_interface]
			    }
			    set theop "_interface"
			    set rtype "Object"
			    set params [list]
			}
			_non_existent {
			    if {[llength $values] != 0} {
				error "usage: $handles_rev($obj) _non_existent"
			    }
			    if {!$async} {
				return [$obj _non_existent]
			    }
			    set theop "_non_existent"
			    set rtype "boolean"
			    set params [list]
			}
			_duplicate {
			    return [corba::string_to_object \
				    [corba::object_to_string \
				    $handles_rev($obj)]]
			}
			default {
			    error "oops"
			}
		    }

		    if {$async} {
			return [invoke_async 1 $obj $theop $rtype 2 \
				$params $values {} $callback]
		    }

		    return [invoke_sync 1 $obj $theop $rtype 2 \
			    $params $values]
		}

		if {$operation == "_is_equivalent"} {
		    if {[llength $values] != 1} {
			error "usage: $handles_rev($obj) _is_equivalent handle"
		    }
		    set otherobj [GetObjFromHandle [lindex $values 0]]
		    if {[catch {
			set res [$obj _is_equivalent $otherobj]
		    } err]} {
			set status 2
			set res $err
		    } else {
			set status 0
		    }
		    if {$async} {
			return [fake_async $status $res $callback]
		    }
		    if {$status == 2} {
			error $res
		    }
		    return $res
		}

		if {![::Combat::SimpleTypeRepository::UpdateTypeInfoForObj $obj]} {
		    ::corba::throw [list IDL:omg.org/CORBA/INTF_REPOS:1.0 \
			    [list minor 0 completion_status COMPLETED_NO]]
		}

		set type [$obj cget -type_id]

		set opInfo [::Combat::SimpleTypeRepository::getOp \
			$type $operation]

		if {$opInfo == ""} {
		    set atInfo [::Combat::SimpleTypeRepository::getAttr \
			    $type $operation]
		} else {
		    set atInfo ""
		}

		if {$opInfo == "" && $atInfo == ""} {
		    catch {$obj UpdateType}

		    if {![::Combat::SimpleTypeRepository::UpdateTypeInfoForObj $obj]} {
			::corba::throw [list IDL:omg.org/CORBA/INTF_REPOS:1.0 \
				[list minor 0 completion_status COMPLETED_NO]]
		    }
		    
		    set type [$obj cget -type_id]
		    set opInfo [::Combat::SimpleTypeRepository::getOp \
			    $type $operation]
		    set atInfo [::Combat::SimpleTypeRepository::getAttr \
			    $type $operation]
		    if {$opInfo == "" && $atInfo == ""} {
			error "error: $operation is neither operation nor attribute for $type"
		    }
		}

		if {$opInfo != ""} {
		    set rtype [lindex $opInfo 9]
		    set params [lindex $opInfo 15]
		    set exceptions [lindex $opInfo 17]

		    if {[llength $params] != [llength $values]} {
			error "error: operation $operation wants [llength $params] parameters, not [llength $values]"
		    }

		    if {[lindex $opInfo 11] == "OP_ONEWAY"} {
			set response 0
		    }

		    if {$async} {
			return [invoke_async $response $obj $operation \
				$rtype 2 $params $values $exceptions \
				$callback]
		    }

		    return [invoke_sync $response $obj $operation \
			    $rtype 2 $params $values $exceptions]
		}

		if {$atInfo != ""} {
		    set attype [lindex $atInfo 9]

		    if {[llength $values] == 0} {
			if {$async} {
			    return [invoke_async 1 $obj _get_$operation \
				    $attype 0 {} {} {} $callback]
			}
			return [invoke_sync 1 $obj _get_$operation \
				$attype 0 {} {}]
		    } elseif {[llength $values] == 1} {
			if {[lindex $atInfo 11] == "ATTR_READONLY"} {
			    error "error: attribute $operation is readonly"
			}
			if {$async} {
			    return [invoke_async 1 $obj _set_$operation \
				    void 0 [list [list in $attype]] $values \
				    {} $callback]
			}

			return [invoke_sync 1 $obj _set_$operation \
				void 0 [list [list in $attype]] $values]
		    }

		    error "attribute get wants 0, attribute set 1 parameter"
		}

		error "oops, I should not be here."
	    }

	    #
	    # corba::dii ?-async? handle spec args
	    #
	    # spec: [list returntype opname params exceptions]
	    #

	    public proc dii_invoke {args} {
		set idx 0
		set async 0
		set callback ""
		set response 1

		if {[lindex $args $idx] == "-async"} {
		    set async 1
		    incr idx
		} elseif {[lindex $args $idx] == "-callback"} {
		    incr idx
		    set async 1
		    set callback [lindex $args $idx]
		    incr idx
		}

		set handle [lindex $args $idx]
		set obj [GetObjFromHandle $handle]
		incr idx

		set spec [lindex $args $idx]
		incr idx

		if {$idx > [llength $args]} {
		    error "too few arguments to dii_invoke"
		}

		if {[llength $spec] < 3} {
		    error "illegal dii call specification"
		}

		set rtype [lindex $spec 0]
		set operation [lindex $spec 1]
		set params [lindex $spec 2]
		set exceptions [lindex $spec 3]
		set values [lrange $args $idx end]

		if {[llength $spec] == 4 && \
			([lindex $spec 3] == "OP_ONEWAY" || \
			[lindex $spec 4] == "oneway")} {
		    set response 0
		}

		if {$async} {
		    return [invoke_async $response $obj $operation \
			    $rtype 2 $params $values $exceptions \
			    $callback]
		}

		return [invoke_sync $response $obj $operation \
			$rtype 2 $params $values $exceptions]
	    }
	}
    }
}
